@@ -46,12 +46,26 @@ import psycopg2
 import requests
 from tqdm import tqdm
 
+from ._camelcase import _decode_camel_case
 from ._csv import _to_csv
 from ._jsonx import _transform_json
+from ._jsonx import _drop_json_tables
 from ._select import _select
-from ._sqlx import _escape_sql
+from ._sqlx import _encode_sql_str
 from ._sqlx import _autocommit
 from ._sqlx import _sqlid
+from ._sqlx import _stage_table
+from ._sqlx import _strip_schema
+from ._sqlx import _varchar_type
+
+def _rename_tables(db, tables):
+    cur = db.cursor()
+    try:
+        for t in tables:
+            cur.execute('DROP TABLE IF EXISTS ' + _sqlid(t))
+            cur.execute('ALTER TABLE ' + _sqlid(_stage_table(t)) + ' RENAME TO ' + _sqlid(_strip_schema(t)))
+    finally:
+        cur.close()
 
 class LDLite:
 
@@ -76,7 +90,7 @@ class LDLite:
         self.page_size = page_size
 
     def connect_db(self, filename):
-        """Connects to an embedded analytic database.
+        """Connects to an embedded database for storing data.
 
         The *filename* specifies a local file containing the database or where
         the database will be created if it does not exist.  By default LDLite
@@ -94,12 +108,11 @@ class LDLite:
         return self.db
 
     def connect_db_postgresql(self, dsn):
-        """Connects to an analytic PostgreSQL database.
+        """Connects to a PostgreSQL database for storing data.
 
-        PostgreSQL can be used as an alternative to the default embedded
-        database.  The data source name is specified by *dsn*.  This function
-        returns a connection to the database which can be used to submit SQL
-        queries.  The returned connection defaults to autocommit mode.
+        The data source name is specified by *dsn*.  This function returns a
+        connection to the database which can be used to submit SQL queries.
+        The returned connection defaults to autocommit mode.
 
         Example:
 
@@ -111,6 +124,23 @@ class LDLite:
         _autocommit(self.db, self.dbtype, True)
         return self.db
 
+    def connect_db_redshift(self, dsn):
+        """Connects to a Redshift database for storing data.
+
+        The data source name is specified by *dsn*.  This function returns a
+        connection to the database which can be used to submit SQL queries.
+        The returned connection defaults to autocommit mode.
+
+        Example:
+
+            db = ld.connect_db_redshift(dsn='dbname=ldlite host=localhost user=ldlite')
+
+        """
+        self.dbtype = 3
+        self.db = psycopg2.connect(dsn)
+        _autocommit(self.db, self.dbtype, True)
+        return self.db
+
     def _login(self):
         if self._verbose:
             print('ldlite: logging in to okapi', file=sys.stderr)
@@ -119,7 +149,12 @@ class LDLite:
         data = { 'username': self.okapi_user,
                 'password': self.okapi_password }
         resp = requests.post(self.okapi_url+'/authn/login', headers=hdr, data=json.dumps(data))
-        self.login_token = resp.headers['x-okapi-token']
+        if resp.status_code != 201:
+            resp.raise_for_status()
+        if 'x-okapi-token' in resp.headers:
+            self.login_token = resp.headers['x-okapi-token']
+        else:
+            raise RuntimeError('authentication service did not return a login token')
 
     def _check_okapi(self):
         if self.login_token is None:
@@ -143,12 +178,45 @@ class LDLite:
                              password='admin')
 
         """
+        if not url.startswith('https://'):
+            raise ValueError('url must begin with "https://"')
         self.okapi_url = url.rstrip('/')
         self.okapi_tenant = tenant
         self.okapi_user = user
         self.okapi_password = password
         self._login()
 
+    def drop_tables(self, table):
+        """Drops a specified table and any accompanying tables that were output from JSON transformation.
+
+        A table called *table*_jtable is used to retrieve the names of the
+        tables created by JSON transformation.
+
+        This function returns a list of all of the dropped tables.
+
+        Example:
+
+            ld.drop_tables('g')
+
+        """
+        schema_table = table.strip().split('.')
+        if len(schema_table) < 1 or len(schema_table) > 2:
+            raise ValueError('invalid table name: ' + table)
+        self._check_db()
+        _autocommit(self.db, self.dbtype, False)
+        try:
+            cur = self.db.cursor()
+            try:
+                cur.execute('DROP TABLE IF EXISTS ' + _sqlid(table))
+            finally:
+                cur.close()
+            tables = [table]
+            tables += _drop_json_tables(self.db, self.dbtype, table)
+            self.db.commit()
+        finally:
+            _autocommit(self.db, self.dbtype, True)
+        return tables
+
     def query(self, table, path, query, json_depth=3, transform=None):
         """Submits a CQL query to an Okapi module, and transforms and stores the result.
 
@@ -175,8 +243,8 @@ class LDLite:
         """
         if transform != None:
             raise ValueError('transform is no longer supported: use json_depth=0 to disable JSON transformation')
-        schema_table = table.strip().split('.')
-        if len(schema_table) < 1 or len(schema_table) > 2:
+        schema_table = table.split('.')
+        if len(schema_table) != 1 and len(schema_table) != 2:
             raise ValueError('invalid table name: ' + table)
         if json_depth is None or json_depth < 0 or json_depth > 4:
             raise ValueError('invalid value for json_depth: ' + str(json_depth))
@@ -185,79 +253,121 @@ class LDLite:
         if not self._quiet:
             print('ldlite: querying: '+path, file=sys.stderr)
         _autocommit(self.db, self.dbtype, False)
-        curout = self.db.cursor()
-        if len(schema_table) == 2:
-            curout.execute('CREATE SCHEMA IF NOT EXISTS ' + _sqlid(schema_table[0]))
-        curout.execute('DROP TABLE IF EXISTS ' + _sqlid(table))
-        curout.execute('CREATE TABLE ' + _sqlid(table) + '(__id integer, jsonb varchar)')
-        # First get total number of records
-        hdr = { 'X-Okapi-Tenant': self.okapi_tenant,
-                'X-Okapi-Token': self.login_token }
-        resp = requests.get(self.okapi_url+path+'?offset=0&limit=1&query='+query, headers=hdr)
-        if resp.status_code == 401:
-            # Retry
-            self._login()
+        try:
+            stage_table = _stage_table(table)
+            cur = self.db.cursor()
+            try:
+                if len(schema_table) == 2:
+                    cur.execute('CREATE SCHEMA IF NOT EXISTS ' + _sqlid(schema_table[0]))
+                cur.execute('DROP TABLE IF EXISTS ' + _sqlid(stage_table))
+                cur.execute('CREATE TABLE ' + _sqlid(stage_table) + '(__id integer, jsonb ' + _varchar_type(self.dbtype) + ')')
+            finally:
+                cur.close()
+            # First get total number of records
             hdr = { 'X-Okapi-Tenant': self.okapi_tenant,
                     'X-Okapi-Token': self.login_token }
             resp = requests.get(self.okapi_url+path+'?offset=0&limit=1&query='+query, headers=hdr)
-        if resp.status_code != 200:
-            resp.raise_for_status()
-        try:
-            j = resp.json()
-        except Exception as e:
-            raise RuntimeError('received server response: ' + resp.text) from e
-        if 'totalRecords' in j:
-            total_records = j['totalRecords']
-        else:
-            total_records = -1
-        total = total_records if total_records is not None else 0
-        if self._verbose:
-            print('ldlite: estimated row count: '+str(total), file=sys.stderr)
-        # Read result pages
-        count = 0
-        page = 0
-        if not self._quiet:
-            if total == -1:
-                pbar = tqdm(desc='reading', leave=False, mininterval=1, smoothing=0, colour='#A9A9A9', bar_format='{desc} {elapsed} {bar}{postfix}')
-            else:
-                pbar = tqdm(desc='reading', total=total, leave=False, mininterval=1, smoothing=0, colour='#A9A9A9', bar_format='{desc} {bar}{postfix}')
-            pbartotal = 0
-        while True:
-            offset = page * self.page_size
-            limit = self.page_size
-            resp = requests.get(self.okapi_url+path+'?offset='+str(offset)+'&limit='+str(limit)+'&query='+query, headers=hdr)
+            if resp.status_code == 401:
+                # Retry
+                self._login()
+                hdr = { 'X-Okapi-Tenant': self.okapi_tenant,
+                        'X-Okapi-Token': self.login_token }
+                resp = requests.get(self.okapi_url+path+'?offset=0&limit=1&query='+query, headers=hdr)
+            if resp.status_code != 200:
+                resp.raise_for_status()
             try:
                 j = resp.json()
             except Exception as e:
                 raise RuntimeError('received server response: ' + resp.text) from e
-            if isinstance(j, dict):
-                data = list(j.values())[0]
+            if 'totalRecords' in j:
+                total_records = j['totalRecords']
             else:
-                data = j
-            lendata = len(data)
-            if lendata == 0:
-                break
-            for d in data:
-                curout.execute('INSERT INTO ' + _sqlid(table) + ' VALUES(' + str(count+1) + ',\'' + _escape_sql(json.dumps(d, indent=4)) + '\')')
-                count += 1
-                if not self._quiet:
-                    if pbartotal + 1 > total:
-                        pbartotal = total
-                        pbar.update(total - pbartotal)
+                total_records = -1
+            total = total_records if total_records is not None else 0
+            if self._verbose:
+                print('ldlite: estimated row count: '+str(total), file=sys.stderr)
+            # Read result pages
+            count = 0
+            page = 0
+            if not self._quiet:
+                if total == -1:
+                    pbar = tqdm(desc='reading', leave=False, mininterval=1, smoothing=0, colour='#A9A9A9', bar_format='{desc} {elapsed} {bar}{postfix}')
+                else:
+                    pbar = tqdm(desc='reading', total=total, leave=False, mininterval=1, smoothing=0, colour='#A9A9A9', bar_format='{desc} {bar}{postfix}')
+                pbartotal = 0
+            cur = self.db.cursor()
+            try:
+                while True:
+                    offset = page * self.page_size
+                    limit = self.page_size
+                    resp = requests.get(self.okapi_url+path+'?offset='+str(offset)+'&limit='+str(limit)+'&query='+query, headers=hdr)
+                    if resp.status_code != 200:
+                        resp.raise_for_status()
+                    try:
+                        j = resp.json()
+                    except Exception as e:
+                        raise RuntimeError('received server response: ' + resp.text) from e
+                    if isinstance(j, dict):
+                        data = list(j.values())[0]
                     else:
+                        data = j
+                    lendata = len(data)
+                    if lendata == 0:
+                        break
+                    for d in data:
+                        cur.execute('INSERT INTO ' + _sqlid(stage_table) + ' VALUES(' + str(count+1) + ',' + _encode_sql_str(self.dbtype, json.dumps(d, indent=4)) + ')')
+                        count += 1
+                        if not self._quiet:
+                            if pbartotal + 1 > total:
+                                pbartotal = total
+                                pbar.update(total - pbartotal)
+                            else:
+                                pbartotal += 1
+                                pbar.update(1)
+                    page += 1
+            finally:
+                cur.close()
+            if not self._quiet:
+                pbar.close()
+            _ = set(_drop_json_tables(self.db, self.dbtype, table))
+            newtables = [table]
+            newattrs = {}
+            if json_depth > 0:
+                jsontables, jsonattrs = _transform_json(self.db, self.dbtype, table, count, self._quiet, json_depth)
+                newtables += jsontables
+                newattrs = jsonattrs
+                for t, attrs in newattrs.items():
+                    newattrs[t]['__id'] = ('__id', 'bigint')
+                newattrs[table] = {'__id': ('__id', 'bigint')}
+            # Rename tables and commit
+            _rename_tables(self.db, newtables)
+            self.db.commit()
+        finally:
+            _autocommit(self.db, self.dbtype, True)
+        # Create indexes
+        if self.dbtype == 2:
+            index_total = sum(map(len, newattrs.values()))
+            if not self._quiet:
+                pbar = tqdm(desc='indexing', total=index_total, leave=False, mininterval=1, smoothing=0, colour='#A9A9A9', bar_format='{desc} {bar}{postfix}')
+                pbartotal = 0
+            for t, attrs in newattrs.items():
+                for attr in sorted(list(attrs)):
+                    decoded_attr, _ = attrs[attr]
+                    cur = self.db.cursor()
+                    try:
+                        cur.execute('CREATE INDEX ON ' + _sqlid(t) + ' (' + _sqlid(decoded_attr) + ')')
+                    except Exception as e:
+                        pass
+                    finally:
+                        cur.close()
+                    if not self._quiet:
                         pbartotal += 1
                         pbar.update(1)
-            page += 1
-        self.db.commit()
-        if not self._quiet:
-            pbar.close()
-        newtables = [table]
-        if json_depth > 0:
-            newtables += _transform_json(self.db, table, count, self._quiet, json_depth, curout)
-        self.db.commit()
+            if not self._quiet:
+                pbar.close()
+        # Return table names
         if not self._quiet:
             print('ldlite: created tables: '+', '.join(newtables), file=sys.stderr)
-        _autocommit(self.db, self.dbtype, True)
         return newtables
 
     def quiet(self, enable):
@@ -292,17 +402,24 @@ class LDLite:
 
         """
         self._check_db()
-        _autocommit(self.db, self.dbtype, True)
         # f = sys.stdout if file is None else file
         f = sys.stdout
         if self._verbose:
             print('ldlite: reading from table: '+table, file=sys.stderr)
-        _select(self.db, table, columns, limit, f)
+        _autocommit(self.db, self.dbtype, False)
+        try:
+            _select(self.db, self.dbtype, table, columns, limit, f)
+            if self.dbtype == 2 or self.dbtype == 3:
+                self.db.rollback()
+        finally:
+            _autocommit(self.db, self.dbtype, True)
 
-    def to_csv(self, filename, table):
+    def to_csv(self, filename, table, header=True):
         """Export a table in the analytic database to a CSV file.
 
-        All rows of *table* are exported to *filename*.
+        All rows of *table* are exported to *filename*.  If *header* is True
+        (the default), the CSV file will begin with a header line containing
+        the column names.
 
         Example:
 
@@ -310,8 +427,13 @@ class LDLite:
 
         """
         self._check_db()
-        _autocommit(self.db, self.dbtype, True)
-        _to_csv(self.db, table, filename)
+        _autocommit(self.db, self.dbtype, False)
+        try:
+            _to_csv(self.db, self.dbtype, table, filename, header)
+            if self.dbtype == 2 or self.dbtype == 3:
+                self.db.rollback()
+        finally:
+            _autocommit(self.db, self.dbtype, True)
 
     def _verbose(self, enable):
         """Configures verbose output.
